/*
 * file: pv_bs.c
 * This file is part of the libpv library.
 *
 * Copyright 2005 Pengutronix
 *
 * Please send bug reports to bugs@pengutronix.de
 *
 * See COPYING for the status of this software.
 */

#include <libpv_config.h>

#include <errno.h>

#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/stat.h>

#include <pv_bs.h>
#include <pv_bs_std.h>
#include <pv_bs_8bit.h>
#include <pv_lock.h>


static int (*pv_bs_store_init_fn[])(struct pv_backing_store	*store,
				    const char			*filename,
				    size_t			size) = {
	[PV_STORE_STANDARD]	= pv_bs_std_init,
	[PV_STORE_8BIT]		= pv_bs_8bit_init,
};


static inline int pv_bs_store_init(struct pv_backing_store	*store,
				   enum pv_bs_store_type	store_type,
				   const char			*store_filename,
				   size_t			size)
{
	return pv_bs_store_init_fn[store_type](store, store_filename, size);
}


static inline int pv_bs_store_destroy(struct pv_backing_store	*store)
{
	return store->exit(store);
}

/**
 * pv_bs_init - Intialize the backing store for this process image
 * @engine: Current engine to work for
 * @store_type: Type of backing store (0=standard, 1=special handling store)
 * @store_filename:
 * @size
 *
 * Description: Initialize the backing store for this process image. If this
 * is the first process it also creates the shared memory and - if requiered -
 * copy the persistant file content into the shared memory image.
 *
 * Return value: 0 in case of success
 **/
int pv_bs_init(struct pv_engine		*engine,
	       enum pv_bs_store_type	store_type,
	       const char		*store_filename,
	       size_t			size)
{
	struct pv_backing_store	*store = &engine->store;
	int			init_store = 0;
	int			err;

	err = pv_bs_store_init(store, store_type, store_filename, size);
	if (err < 0) {
		dprintf(LEVEL_CRIT, "(%s) could not initialize backing store, err %d\n",
			__FUNCTION__, err);
		goto exit_0;
	}

	if (!store->map) {
	/*
	 * Process image is opened on the backing
	 * store but it is not mmap-RW capable.
	 * In this case create an additional shared
	 * memory area as a working copy
	 */
		dprintf(LEVEL_DEBUG, "(%s) opening shared memory segment\n",
			__FUNCTION__);

		pv_lock_lock(engine);

		store->shmid = shmget(engine->pv_sem_key, size,
				      IPC_CREAT | IPC_EXCL | S_IRUSR | S_IWUSR );

		if(store->shmid < 0 && errno == EEXIST) {	/* already existing */
			dprintf(LEVEL_INFO, "(%s) shared memory segment already exists\n",
				__FUNCTION__);

			store->shmid = shmget(engine->pv_sem_key, size,
					      IPC_CREAT | S_IRUSR | S_IWUSR);
			if(store->shmid < 0) {
				err = -errno;
				dprintf(LEVEL_CRIT, "(%s) failed to get shared memory segment id, key 0x%08x, err %d\n",
					__FUNCTION__, engine->pv_sem_key, err);
				goto exit_1;
			}
		} else {
			dprintf(LEVEL_INFO, "(%s) creating shared memory segment\n",
				__FUNCTION__);
			init_store = 1;
		}

		store->map = shmat(store->shmid, NULL, 0);
		if (store->map == (void*)-1) {
			err = -errno;
			dprintf(LEVEL_CRIT, "(%s) failed to attach to shared memory segment, err %d\n",
				__FUNCTION__, err);
			goto exit_1;
		}
		dprintf(LEVEL_INFO, "(%s) attached to shared memory segment key 0x%08x, id %d\n",
			__FUNCTION__, engine->pv_sem_key, store->shmid);
		/*
		 * As we will working on a copy we first have to read back
		 * the current state from backing store
		 */
		if(init_store) {
			dprintf(LEVEL_INFO, "(%s) initialising store\n",
				__FUNCTION__);

			store->read(store, 0, store->size);
		}
		pv_lock_unlock(engine);
	} else
		store->shmid = -1;	/* no additional shared memory required */

	return 0;

 exit_1:
	pv_lock_unlock(engine);
 exit_0:
	return err;
}

/**
 * pv_bs_destroy - destroy all claimed resources for this process image
 * @engine:
 *
 * Description:
 *
 * Return value:
 **/
int pv_bs_destroy(struct pv_engine	*engine)
{
	struct pv_backing_store	*store = &engine->store;
	int			err = 0, ret;

	ret = pv_bs_store_destroy(store);
	if (ret) {
		err = ret;
		dprintf(LEVEL_ERROR, "(%s) failed to destroy backing store, err %d\n",
			__FUNCTION__, err);
	}

	if (store->shmid != -1) {
		dprintf(LEVEL_DEBUG, "(%s) detaching shared memory\n", __FUNCTION__);
		shmdt(engine->store.map);
		shmctl(store->shmid, IPC_RMID, NULL);
	}

	return err;
}

/* end of file pv_bs.c */
